use std::process::{Command, Stdio};

use based::get_pg;

use crate::{
    git::get_commit_hash,
    pacco_push,
    package::{PackageIndex, find_pkgs, get_pkgver},
};

pub struct BuildEnv(String, bool);

impl BuildEnv {
    pub fn new() -> Self {
        let tmp_env = format!("./build/pkg-{}", uuid::Uuid::new_v4().to_string());

        Self::setup_root(&tmp_env);

        Self(tmp_env, true)
    }

    pub fn setup_root(path: &str) {
        if !std::fs::exists(std::path::Path::new(path)).unwrap() {
            log::info!("Setting up root fs at {path}");
            let _ = Command::new("mkdir").args(["-p", path]).status();
            let _ = Command::new("pacstrap")
                .args(["-c", path, "base", "base-devel"])
                .status();
        } else {
            log::info!("Reusing existing root fs at {path}");
        }
    }

    pub fn build_dir(&self) -> String {
        std::path::Path::new(&self.0)
            .join("build")
            .to_str()
            .unwrap()
            .to_string()
    }

    pub fn new_from(path: &str) -> Self {
        Self::setup_root(path);

        Self(path.to_owned(), false)
    }

    pub fn path(&self) -> &str {
        self.0.as_str()
    }

    pub fn copy_build_env_dir(&self, dir: &str) {
        log::info!(
            "Copying build env from {dir} -> {}",
            std::path::Path::new(&self.0)
                .join("build")
                .to_str()
                .unwrap()
        );
        Command::new("rsync")
            .arg("-azhru")
            .arg("--delete")
            .arg(format!("{dir}/"))
            .arg(format!(
                "{}/",
                std::path::Path::new(&self.0)
                    .join("build")
                    .to_str()
                    .unwrap()
            ))
            .output()
            .unwrap();
    }

    pub fn run_command_inherit(&self, cmd: &str) -> bool {
        let output = Command::new("systemd-nspawn")
            .arg("-D")
            .arg(self.path())
            .arg("--bind")
            .arg("./pacman.conf:/etc/pacman.conf")
            .arg("bash")
            .arg("-c")
            .arg(cmd)
            .stdout(Stdio::inherit())
            .output()
            .expect("Failed to run systemd-nspawn container");

        let success = output.status.success();

        if !success {
            let stderr = String::from_utf8_lossy(&output.stderr).to_string();
            log::error!("Error running command: {stderr}");
        }

        success
    }

    pub fn run_command(&self, cmd: &str) -> (bool, String) {
        let output = Command::new("systemd-nspawn")
            .arg("-D")
            .arg(self.path())
            .arg("--bind")
            .arg("./pacman.conf:/etc/pacman.conf")
            .arg("bash")
            .arg("-c")
            .arg(cmd)
            .output()
            .expect("Failed to run systemd-nspawn container");

        let stdout = String::from_utf8_lossy(&output.stdout).to_string();
        let success = output.status.success();

        if !success {
            let stderr = String::from_utf8_lossy(&output.stderr).to_string();
            log::error!("Error running command: {stderr}\nStdout: {stdout}");
        }

        (success, stdout)
    }

    pub fn cleanup(&self) {
        if self.1 {
            Command::new("umount")
                .arg(std::path::Path::new(&self.0).join("proc"))
                .output()
                .unwrap();
            log::info!("Removing temporary build environment at {}", self.0);
            std::fs::remove_dir_all(&self.0).unwrap();
        }
    }
}

impl Drop for BuildEnv {
    fn drop(&mut self) {
        self.cleanup();
    }
}

pub async fn build(pkg: &PackageIndex) -> (String, Vec<(String, Vec<u8>)>) {
    let base_env = BuildEnv::new();

    let commit = get_commit_hash(pkg.path().to_str().unwrap());
    let pkgver = get_pkgver(pkg.path().to_str().unwrap());

    base_env.copy_build_env_dir(&pkg.build_dir());

    log::info!(
        "Building {} / {} @ {}{}",
        pkg.repo,
        pkg.pkg,
        pkgver.as_ref().unwrap_or(&String::new()),
        if let Some(c) = &commit {
            format!(" [#{c}]")
        } else {
            String::new()
        }
    );

    let success = base_env.run_command_inherit(
        r#"cd /build;pacman -Syu --noconfirm;useradd -m build;echo "ALL ALL=(ALL) NOPASSWD: ALL"|tee -a /etc/sudoers;chown -R build /build;su -c "makepkg -c -C -s --noconfirm --skippgpcheck" build"#);

    if success {
        sqlx::query(
            "UPDATE package SET last_commit = $1, last_pkgver = $2 WHERE repo = $3 AND pkg = $4",
        )
        .bind(&commit)
        .bind(&pkgver)
        .bind(&pkg.repo)
        .bind(&pkg.pkg)
        .execute(get_pg!())
        .await
        .unwrap();

        let packages = find_pkgs(&base_env.build_dir());

        for package in &packages {
            log::info!(
                "Successfully built {} / {} @ {}{}",
                pkg.repo,
                pkg.pkg,
                pkgver.as_ref().unwrap_or(&String::new()),
                if let Some(c) = &commit {
                    format!(" [#{c}]")
                } else {
                    String::new()
                }
            );

            std::fs::create_dir_all(std::path::Path::new("./packages").join(&pkg.repo)).unwrap();
            std::fs::write(
                std::path::Path::new("./packages")
                    .join(&pkg.repo)
                    .join(&package.0),
                &package.1,
            )
            .unwrap();

            pacco_push(
                std::path::Path::new("./packages")
                    .join(&pkg.repo)
                    .join(&package.0)
                    .to_str()
                    .unwrap(),
                &pkg.repo,
            )
            .await;
        }

        (String::new(), packages)
    } else {
        log::error!(
            "Error building {} / {} @ {}{}",
            pkg.repo,
            pkg.pkg,
            pkgver.as_ref().unwrap_or(&String::new()),
            if let Some(c) = &commit {
                format!(" [#{c}]")
            } else {
                String::new()
            }
        );
        (String::new(), Vec::new())
    }
}
