use based::get_pg;
use serde::{Deserialize, Serialize};
use sqlx::FromRow;
use std::fs::read_dir;
use std::io::Read;
use std::path::Path;
use std::{fs, path::PathBuf};

use crate::builder::{BuildEnv, build};
use crate::git::{is_git, pull};

#[derive(Deserialize, Serialize, FromRow)]
pub struct PackageIndex {
    pub repo: String,
    pub pkg: String,
    pub last_commit: Option<String>,
    pub last_pkgver: Option<String>,
}

impl PackageIndex {
    pub async fn all() -> Vec<Self> {
        sqlx::query_as("SELECT * FROM package")
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    pub fn path(&self) -> PathBuf {
        std::path::Path::new("./repositories")
            .join(&self.repo)
            .join(&self.pkg)
    }

    pub fn build_dir(&self) -> String {
        format!("./repositories/{}/{}", self.repo, self.pkg)
    }
}

pub async fn rebuild_pkgs() {
    log::info!("Start checking for package rebuilds");
    let all_pkgs = PackageIndex::all().await;

    for pkg in &all_pkgs {
        let mut should_rebuild = false;

        if is_git(pkg.path().to_str().unwrap()) {
            let git_commit = pull(pkg.path().to_str().unwrap()).unwrap_or_default();
            if let Some(old_commit) = &pkg.last_commit {
                if *old_commit != git_commit {
                    log::info!(
                        "Package {} / {} should rebuild: Commit {} -> {}",
                        pkg.repo,
                        pkg.pkg,
                        old_commit,
                        git_commit
                    );
                    should_rebuild = true;
                }
            } else {
                log::info!(
                    "Package {} / {} should rebuild: No last commit",
                    pkg.repo,
                    pkg.pkg
                );
                should_rebuild = true;
            }
        }

        let pkgver = get_pkgver(pkg.path().to_str().unwrap()).unwrap_or_default();

        if let Some(old_pkgver) = &pkg.last_pkgver {
            if *old_pkgver != pkgver {
                log::info!(
                    "Package {} / {} should rebuild: pkgver {} -> {}",
                    pkg.repo,
                    pkg.pkg,
                    old_pkgver,
                    pkgver
                );
                should_rebuild = true;
            }
        } else {
            log::info!(
                "Package {} / {} should rebuild: No last pkgver",
                pkg.repo,
                pkg.pkg
            );
            should_rebuild = true;
        }

        if should_rebuild {
            build(pkg).await;
        }
    }
    log::info!("Done checking for package rebuilds");
}

pub fn get_pkgver(repo: &str) -> Option<String> {
    // let base_env = BuildEnv::new_from("./build/srcinfo");
    let base_env = BuildEnv::new();

    base_env.copy_build_env_dir(repo);

    let (success, out) = base_env.run_command(
        r#"cd /build;useradd -m build;echo "ALL ALL=(ALL) NOPASSWD: ALL"|tee -a /etc/sudoers;chown -R build /build;su -c "makepkg --printsrcinfo" build"#);

    if success {
        for line in out.lines() {
            let trimmed = line.trim();
            if let Some(pkgver) = trimmed.strip_prefix("pkgver = ") {
                return Some(pkgver.to_string());
            }
        }
    }

    None
}

pub fn find_pkgs(dir: &str) -> Vec<(String, Vec<u8>)> {
    let mut pkgs = Vec::new();

    let path = Path::new(dir);

    if let Ok(entries) = fs::read_dir(path) {
        for entry in entries.flatten() {
            let file_path = entry.path();
            if let Some(ext) = file_path.extension() {
                if ext == "zst" {
                    if let Some(file_name) = file_path.file_name().and_then(|n| n.to_str()) {
                        if file_name.ends_with(".tar.zst") {
                            match fs::File::open(&file_path) {
                                Ok(mut file) => {
                                    let mut content = Vec::new();
                                    if file.read_to_end(&mut content).is_ok() {
                                        pkgs.push((file_name.to_string(), content));
                                    }
                                }
                                Err(_) => continue,
                            }
                        }
                    }
                }
            }
        }
    }

    pkgs
}

pub async fn reindex_pkg(repo_dir: &Path) {
    log::info!("Indexing package builds");
    let mut dir_entries = read_dir(repo_dir).unwrap();

    while let Some(Ok(repo_entry)) = dir_entries.next() {
        if repo_entry.file_type().unwrap().is_dir() {
            let repo_name = repo_entry.file_name().into_string().unwrap();
            let mut pkg_entries = read_dir(repo_entry.path()).unwrap();

            while let Some(Ok(pkg_entry)) = pkg_entries.next() {
                if pkg_entry.file_type().unwrap().is_dir() {
                    let pkg_name = pkg_entry.file_name().into_string().unwrap();
                    log::info!("Found package {repo_name} / {pkg_name}");

                    sqlx::query(
                        "INSERT INTO package (repo, pkg) VALUES ($1, $2) ON CONFLICT DO NOTHING",
                    )
                    .bind(&repo_name)
                    .bind(&pkg_name)
                    .execute(get_pg!())
                    .await
                    .unwrap();
                }
            }
        }
    }

    log::info!("Done indexing package builds");
}
